#!/usr/bin/env node

var fs = require('fs');
var path = require('path');
var mkdirp = require('mkdirp');
var hyperquest = require('hyperquest');
var concat = require('concat-stream');
var rsa = require('hybrid-rsa-stream');
var level = require('level');
var defined = require('defined');
var inherits = require('inherits');
var EventEmitter = require('events').EventEmitter;
var through = require('through2');
var readonly = require('read-only-stream');

var HOME = process.env && (process.env.HOME || process.env.USERPROFILE);

module.exports = Cipherhub;
inherits(Cipherhub, EventEmitter);

function Cipherhub (opts) {
    var self = this;
    if (!(self instanceof Cipherhub)) return new Cipherhub(opts);
    if (!opts) opts = {};
    
    var dir = defined(opts.dir, path.join(HOME, '.config', 'cipherhub'));
    
    mkdirp(dir, function (err) {
        self.db = level(path.join(dir, 'keys.db'));
        self.emit('_ready');
    });
}

Cipherhub.prototype.add = function (user, cb) {
    var self = this;
    return concat(function f (body) {
        if (!self.db) {
            return self.once('_ready', function () { f(body) });
        }
        self.db.put(user, body.toString('utf8'), function (err) {
            if (err) cb(err)
            else cb(null, body)
        });
    });
};

Cipherhub.prototype.list = function f () {
    var self = this;
    if (!self.db) {
        var out = through();
        self.once('_ready', function () {
            f().pipe(out);
        });
        return readonly(out);
    }
    return self.db.createReadStream();
};

Cipherhub.prototype.remove = function f (user, cb) {
    var self = this;
    if (!self.db) {
        return self.once('_ready', function () { f(cb) });
    }
    return self.db.del(user, cb);
};

Cipherhub.prototype.decrypt = function () {
    var keyfile = argv.decrypt;
    if (keyfile === true) {
        keyfile = path.join(
            process.env.HOME || process.env.USERPROFILE,
            '.ssh', 'id_rsa'
        );
    }
    return fs.readFile(keyfile, function (err, privkey) {
        if (err) {
            console.error(err);
            process.exit(1);
        }
        process.stdin
            .pipe(rsa.decrypt(privkey, { encoding: argv.encoding }))
            .on('error', onError)
            .pipe(process.stdout)
        ;

        function onError (err) {
            console.error('cipherhub: you tried to decrypt something')
            console.error('           that was not a message for you!')
            throw err
        }
    });
}


if (argv._.length === 0) {
    fs.createReadStream(__dirname + '/usage.txt').pipe(process.stdout);
    return;
}


var user = argv._[0];
db = getDb();
keyOf(user, function (err, keys, fromGithub) {
    if (err) {
        console.error(err);
        return process.exit(10);
    }
    if (!keys || keys.length === 0) {
        console.error(
            'No RSA keys available for the requested user.\n'
            + 'Add a key manually by doing:\n\n'
            + '  cipherhub --add ' + user + ' < rsa.pub'
            + '\n'
        );
        return process.exit(20);
    }
    if (keys.length > 1) {
        console.error(
            'Multiple keys available for the user: ' + user + ':\n\n'
            + keys.map(function (key) {
                return '  ' + key;
            }).join('\n') + '\n\n'
            + 'Add a key manually by doing:\n\n'
            + ' cipherhub --add ' + user + ' <<< KEYDATA\n'
            + '\n'
        );
        return process.exit(21);
    }
    
    if (fromGithub && argv.save !== false) {
        db.put(user, keys[0], function (err) {
            if (err) {
                console.error('Error saving key for user', user + ':');
                console.error(err);
            }
            else if (!argv.quiet) {
                console.error('# saved new key for user ', user);
                encrypt();
            }
            else encrypt();
        });
    }
    else encrypt();
    
    function encrypt () {
        var enc = rsa.encrypt(keys[0], { encoding: argv.encoding });
        process.stdin.pipe(enc).pipe(process.stdout);
        enc.on('end', function () { process.stdout.write('\n') });
    }
});

function keyOf (user, cb) {
    db.get(user, function (err, row) {
        if (err && err.name === 'NotFoundError') {
            if (argv.github === false) {
                cb(null, undefined);
            }
            else githubKeys(user, function (err, keys) {
                cb(err, keys, true);
            })
        }
        else if (err) cb(err)
        else cb(null, [row])
    });
}

function githubKeys (user, cb) {
    var u = 'https://github.com/' + user + '.keys';
    var hq = hyperquest(u);
    hq.on('error', function (err) {
        cb(err);
        cb = function () {};
    });
    hq.pipe(concat(function (body) {
        var keys = body.toString().split(/\r?\n/)
            .map(function (key) { return key.trim() })
            .filter(function (key) {
                return /^ssh-rsa\b/.test(key);
            })
        ;
        cb(null, keys);
    }));
}
